"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _root, _canvasElement, _canvas, _isDragging, _newTransition, _lastPosX, _lastPosY, _initialState, _currentState, _transitions, _statesMap, _backgroundColor, _maxZoom, _minZoom;
Object.defineProperty(exports, "__esModule", { value: true });
const fabric_1 = require("fabric");
const state_1 = __importDefault(require("./state"));
const transition_1 = __importDefault(require("./transition"));
const common_1 = require("./common");
class StateMachineEditor extends HTMLElement {
    constructor() {
        super();
        // html and canvas elements
        _root.set(this, void 0);
        _canvasElement.set(this, void 0);
        _canvas.set(this, void 0);
        // internal state
        _isDragging.set(this, false);
        _newTransition.set(this, null);
        _lastPosX.set(this, 0);
        _lastPosY.set(this, 0);
        this.stateMachineRules = {};
        _initialState.set(this, '');
        _currentState.set(this, '');
        _transitions.set(this, []);
        _statesMap.set(this, {});
        // configuration
        _backgroundColor.set(this, common_1.EDITOR_DEFAULTS.backgroundColor);
        _maxZoom.set(this, common_1.EDITOR_DEFAULTS.maxZoom);
        _minZoom.set(this, common_1.EDITOR_DEFAULTS.minZoom);
        __classPrivateFieldSet(this, _root, this.attachShadow({ mode: 'open' }));
        __classPrivateFieldSet(this, _canvasElement, document.createElement('canvas'));
        __classPrivateFieldGet(this, _root).appendChild(__classPrivateFieldGet(this, _canvasElement));
        __classPrivateFieldSet(this, _canvas, __classPrivateFieldSet(this, _canvas, new fabric_1.fabric.Canvas(__classPrivateFieldGet(this, _canvasElement), {
            backgroundColor: __classPrivateFieldGet(this, _backgroundColor).toString(),
            selection: false,
            fireRightClick: true,
            stopContextMenu: true,
        })));
        __classPrivateFieldGet(this, _canvas).on('mouse:down', this.handleMouseDown.bind(this));
        __classPrivateFieldGet(this, _canvas).on('mouse:up', this.handleMouseUp.bind(this));
        __classPrivateFieldGet(this, _canvas).on('mouse:wheel', this.handleMouseWheel.bind(this));
        __classPrivateFieldGet(this, _canvas).on('mouse:move', this.handleMouseMove.bind(this));
        __classPrivateFieldGet(this, _canvas).on('mouse:dblclick', this.handleDoubleClick.bind(this));
    }
    connectedCallback() {
        this.setCanvasSize(Number(this.getAttribute('width')) || 100, Number(this.getAttribute('height')) || 100);
    }
    get stateMachine() {
        return this.getStateMachine();
    }
    set stateMachine(value) {
        this.setStateMachine(value);
    }
    get currentState() {
        return this.getAttribute('currentstate');
    }
    set currentState(value) {
        if (value) {
            this.setAttribute('currentstate', value);
        }
    }
    get width() {
        return this.getAttribute('width');
    }
    set width(newWidth) {
        if (newWidth == null) {
            return;
        }
        this.setWidth(Number(newWidth));
        this.setAttribute('width', newWidth);
    }
    get height() {
        return this.getAttribute('height');
    }
    set height(newHeight) {
        if (newHeight == null)
            return;
        this.setHeight(Number(newHeight));
        this.setAttribute('height', newHeight);
    }
    static get observedAttributes() {
        return ['width', 'height', 'currentstate'];
    }
    attributeChangedCallback(name, _oldValue, newValue) {
        switch (name) {
            case 'width':
                this.setWidth(Number(newValue));
                break;
            case 'height':
                this.setHeight(Number(newValue));
                break;
            case 'currentstate':
                this.setCurrentState(newValue);
                break;
            default:
                console.log('defalut case');
                break;
        }
    }
    initialRender() {
        // clear old render
        __classPrivateFieldSet(this, _transitions, []);
        __classPrivateFieldSet(this, _statesMap, {});
        __classPrivateFieldGet(this, _canvas).clear();
        __classPrivateFieldGet(this, _canvas).setBackgroundColor(__classPrivateFieldGet(this, _backgroundColor).toString(), () => {
            console.log('cleaerd');
        });
        // new render
        const stateList = Object.entries(this.stateMachineRules);
        stateList.forEach(([stateName, { metadata: { x, y }, },]) => {
            this.drawState(stateName, x, y);
        });
        stateList.forEach(([stateName, { events }]) => {
            const eventKeys = Object.keys(events);
            const src = __classPrivateFieldGet(this, _statesMap)[stateName];
            eventKeys.forEach((event) => {
                const destStateName = events[event];
                const dest = __classPrivateFieldGet(this, _statesMap)[destStateName];
                this.drawTransition(event, { src, dest });
            });
        });
    }
    // ACTIONS
    setCurrentState(value) {
        const oldState = __classPrivateFieldGet(this, _statesMap)[__classPrivateFieldGet(this, _currentState)];
        if (oldState) {
            oldState.active = false;
        }
        const newState = __classPrivateFieldGet(this, _statesMap)[value];
        if (newState) {
            newState.active = true;
            __classPrivateFieldSet(this, _currentState, value);
        }
    }
    setStateMachine({ initialState, rules }) {
        this.stateMachineRules = rules;
        __classPrivateFieldSet(this, _initialState, initialState);
        this.initialRender();
        this.setCurrentState(initialState);
    }
    addState(name, x, y) {
        if (typeof name !== 'string' || this.stateMachineRules[name] != null) {
            throw new Error('State Already Exists');
        }
        this.stateMachineRules[name] = {
            metadata: { x, y },
            events: {},
        };
        this.drawState(name, x, y);
        this.handleStateMachineChange();
    }
    setInitialState(name) {
        const newInitialState = __classPrivateFieldGet(this, _statesMap)[name];
        const oldInitialState = __classPrivateFieldGet(this, _statesMap)[__classPrivateFieldGet(this, _initialState)];
        if (newInitialState)
            newInitialState.initial = true;
        if (oldInitialState)
            oldInitialState.initial = false;
        __classPrivateFieldSet(this, _initialState, name);
    }
    deleteState(name) {
        if (__classPrivateFieldGet(this, _statesMap)[name] == null) {
            throw new Error('Invalid State Item');
        }
        // clear state instance
        __classPrivateFieldGet(this, _statesMap)[name].remove();
        delete __classPrivateFieldGet(this, _statesMap)[name];
        // clear transition instances
        __classPrivateFieldSet(this, _transitions, __classPrivateFieldGet(this, _transitions).filter((transition) => {
            if (transition.src?.name === name || transition.dest?.name === name) {
                transition.remove();
                return false;
            }
            return true;
        }));
        this.handleStateMachineChange();
    }
    deleteTransition(id) {
        __classPrivateFieldSet(this, _transitions, __classPrivateFieldGet(this, _transitions).filter((transition) => {
            if (transition.id === id && transition.src) {
                transition.remove();
                return false;
            }
            return true;
        }));
        this.handleStateMachineChange();
    }
    startAddTransition(event, srcName, destX, destY) {
        if (__classPrivateFieldGet(this, _statesMap)[srcName] == null) {
            throw new Error('Invalid State Item');
        }
        if (__classPrivateFieldGet(this, _newTransition) != null) {
            throw new Error('Already adding transition');
        }
        const src = __classPrivateFieldGet(this, _statesMap)[srcName];
        __classPrivateFieldSet(this, _newTransition, this.drawTransition(event, {
            src,
            defaultX: destX,
            defaultY: destY,
        }));
    }
    stopAddTransition() {
        if (__classPrivateFieldGet(this, _newTransition) == null)
            return;
        __classPrivateFieldGet(this, _newTransition).remove();
        __classPrivateFieldSet(this, _newTransition, null);
    }
    setCanvasSize(width, height) {
        __classPrivateFieldGet(this, _canvas).setWidth(width);
        __classPrivateFieldGet(this, _canvas).setHeight(height);
    }
    setHeight(height) {
        __classPrivateFieldGet(this, _canvas).setHeight(height);
    }
    setWidth(width) {
        __classPrivateFieldGet(this, _canvas).setWidth(width);
    }
    setHoverCursor(cursor) {
        __classPrivateFieldGet(this, _canvas).hoverCursor = cursor;
    }
    getStateMachine() {
        const newDef = {};
        for (const state of Object.values(__classPrivateFieldGet(this, _statesMap))) {
            newDef[state.name] = { metadata: state.position, events: {} };
        }
        __classPrivateFieldGet(this, _transitions).forEach((transition) => {
            if (transition.src && transition.dest) {
                newDef[transition.src.name].events[transition.name] = transition.dest.name;
            }
        });
        this.stateMachineRules = newDef;
        return JSON.parse(JSON.stringify(this.stateMachineRules));
    }
    // CANVAS EVENT HANDLERS
    dispatchCustomEvent(name, detail) {
        this.dispatchEvent(new CustomEvent(name, {
            detail,
        }));
    }
    updateAllObjects() {
        __classPrivateFieldGet(this, _canvas).requestRenderAll();
        Object.values(__classPrivateFieldGet(this, _statesMap)).forEach((state) => state.setCoords());
        __classPrivateFieldGet(this, _transitions).forEach((transition) => {
            transition.setCoords();
        });
    }
    handleMouseWheel({ e }) {
        e.preventDefault();
        e.stopPropagation();
        const { offsetX, offsetY, deltaY } = e;
        let zoom = __classPrivateFieldGet(this, _canvas).getZoom() + deltaY / 200;
        if (zoom > __classPrivateFieldGet(this, _maxZoom))
            zoom = __classPrivateFieldGet(this, _maxZoom);
        if (zoom < __classPrivateFieldGet(this, _minZoom))
            zoom = __classPrivateFieldGet(this, _minZoom);
        __classPrivateFieldGet(this, _canvas).zoomToPoint(new fabric_1.fabric.Point(offsetX, offsetY), zoom);
        this.updateAllObjects();
    }
    handleMouseDown({ button, target, e, absolutePointer }) {
        const { altKey, clientX, clientY } = e;
        if (target != null)
            return;
        if (altKey) {
            __classPrivateFieldSet(this, _isDragging, true);
            __classPrivateFieldSet(this, _lastPosX, clientX);
            __classPrivateFieldSet(this, _lastPosY, clientY);
            return;
        }
        const detail = {
            event: e,
            canvas: { x: absolutePointer?.x, y: absolutePointer?.y },
        };
        if (button === 3) {
            this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.canvas.rightClick, detail);
        }
        else {
            this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.canvas.click, detail);
        }
    }
    handleMouseUp() {
        __classPrivateFieldSet(this, _isDragging, false);
    }
    handleMouseMove({ absolutePointer, e }) {
        if (__classPrivateFieldGet(this, _isDragging)) {
            const { clientX, clientY } = e;
            if (__classPrivateFieldGet(this, _canvas).viewportTransform &&
                __classPrivateFieldGet(this, _canvas).viewportTransform[4] &&
                __classPrivateFieldGet(this, _canvas).viewportTransform[5]) {
                __classPrivateFieldGet(this, _canvas).viewportTransform[4] += clientX - __classPrivateFieldGet(this, _lastPosX);
                __classPrivateFieldGet(this, _canvas).viewportTransform[5] += clientY - __classPrivateFieldGet(this, _lastPosY);
            }
            __classPrivateFieldSet(this, _lastPosX, clientX);
            __classPrivateFieldSet(this, _lastPosY, clientY);
            this.updateAllObjects();
        }
        // update new transition destination
        if (__classPrivateFieldGet(this, _newTransition) == null || absolutePointer == null)
            return;
        if (__classPrivateFieldGet(this, _newTransition).src && __classPrivateFieldGet(this, _newTransition).dest == null) {
            __classPrivateFieldGet(this, _newTransition).defaultDest = {
                x: absolutePointer.x,
                y: absolutePointer.y,
            };
        }
    }
    handleDoubleClick({ absolutePointer, target, e: event }) {
        if (target != null)
            return;
        this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.canvas.doubleClick, {
            canvas: { x: absolutePointer?.x, y: absolutePointer?.y },
            event: event,
        });
    }
    // STATE EVENTS
    handleStateMachineChange() {
        this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.stateMachineChange, {
            stateMachine: this.getStateMachine(),
        });
    }
    handleStateMove(_opt, state) {
        __classPrivateFieldGet(this, _transitions).forEach((transition) => {
            if (transition.dest && transition.dest.name === state.name) {
                transition.dest = state;
            }
            if (transition.src && transition.src.name === state.name) {
                transition.src = state;
            }
            transition.setCoords();
        });
        state.setCoords();
    }
    handleStateDoubleClick({ absolutePointer, e: event }, state) {
        const { name, position: { x, y }, } = state;
        this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.state.doubleClick, {
            state: { name, x, y },
            canvas: { x: absolutePointer?.x, y: absolutePointer?.y },
            event: event,
        });
    }
    handleStateMouseOver({ e: event }, state) {
        const { position: { x, y }, name, } = state;
        this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.state.over, {
            state: { name, x, y },
            event: event,
        });
        if (__classPrivateFieldGet(this, _newTransition)) {
            __classPrivateFieldGet(this, _newTransition).dest = state;
        }
    }
    handleStateMouseOut({ e }, state) {
        const { position: { x, y }, name, } = state;
        this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.state.out, {
            state: { name, x, y },
            event: e,
        });
        if (__classPrivateFieldGet(this, _newTransition) == null)
            return;
        if (__classPrivateFieldGet(this, _newTransition).dest) {
            __classPrivateFieldGet(this, _newTransition).dest = null;
        }
    }
    handleStateClick({ button, e }, state) {
        const { position: { x, y }, name, } = state;
        const detail = {
            state: { name, x, y },
            event: e,
        };
        if (button === 3) {
            this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.state.rightClick, detail);
        }
        else {
            this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.state.click, detail);
        }
        if (__classPrivateFieldGet(this, _newTransition) == null)
            return;
        if (__classPrivateFieldGet(this, _newTransition).src && __classPrivateFieldGet(this, _newTransition).dest) {
            this.stateMachineRules[__classPrivateFieldGet(this, _newTransition).src.name].events[__classPrivateFieldGet(this, _newTransition).name] = __classPrivateFieldGet(this, _newTransition).dest.name;
            __classPrivateFieldGet(this, _transitions).push(__classPrivateFieldGet(this, _newTransition));
            __classPrivateFieldSet(this, _newTransition, null);
            this.handleStateMachineChange();
        }
    }
    handleStateMouseUp() {
        this.handleStateMachineChange();
    }
    // TRANSITION
    handleTransitionClick({ button, e }, transition) {
        const detail = {
            transition,
            event: e,
        };
        if (button === 3) {
            this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.transition.rightClick, detail);
        }
        else {
            this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.transition.click, detail);
        }
    }
    handleTransitionOver({ e: event }, transition) {
        const detail = {
            transition,
            event: event,
        };
        this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.transition.over, detail);
    }
    handleTransitionOut({ e: event }, transition) {
        const detail = {
            transition,
            event: event,
        };
        this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.transition.over, detail);
    }
    handleTransitionDoubleClick({ e: event }, transition) {
        const detail = {
            transition,
            event: event,
        };
        this.dispatchCustomEvent(common_1.ELEMENT_EVENTS.transition.doubleClick, detail);
    }
    // DRAW METHODS
    drawState(name, x, y) {
        const state = new state_1.default(__classPrivateFieldGet(this, _canvas), name, x, y, {
            initial: __classPrivateFieldGet(this, _initialState) === name,
            active: __classPrivateFieldGet(this, _currentState) === name,
        });
        state.on(common_1.ObjectEvent.MouseOut, this.handleStateMouseOut.bind(this));
        state.on(common_1.ObjectEvent.MouseOver, this.handleStateMouseOver.bind(this));
        state.on(common_1.ObjectEvent.MouseDown, this.handleStateClick.bind(this));
        state.on(common_1.ObjectEvent.MouseUp, this.handleStateMouseUp.bind(this));
        state.on(common_1.ObjectEvent.MouseDoubleClick, this.handleStateDoubleClick.bind(this));
        state.on(common_1.ObjectEvent.Moving, this.handleStateMove.bind(this));
        state.render();
        __classPrivateFieldGet(this, _statesMap)[name] = state;
        return state;
    }
    drawTransition(event, options) {
        const transition = new transition_1.default(__classPrivateFieldGet(this, _canvas), event, options);
        transition.on(common_1.ObjectEvent.MouseOver, this.handleTransitionOver.bind(this));
        transition.on(common_1.ObjectEvent.MouseOut, this.handleTransitionOut.bind(this));
        transition.on(common_1.ObjectEvent.MouseDown, this.handleTransitionClick.bind(this));
        transition.on(common_1.ObjectEvent.MouseDoubleClick, this.handleTransitionDoubleClick.bind(this));
        __classPrivateFieldGet(this, _transitions).push(transition);
        return transition;
    }
}
exports.default = StateMachineEditor;
_root = new WeakMap(), _canvasElement = new WeakMap(), _canvas = new WeakMap(), _isDragging = new WeakMap(), _newTransition = new WeakMap(), _lastPosX = new WeakMap(), _lastPosY = new WeakMap(), _initialState = new WeakMap(), _currentState = new WeakMap(), _transitions = new WeakMap(), _statesMap = new WeakMap(), _backgroundColor = new WeakMap(), _maxZoom = new WeakMap(), _minZoom = new WeakMap();
if (!window.customElements.get('state-machine-editor')) {
    window.customElements.define('state-machine-editor', StateMachineEditor);
}
// new StateMachineEditor("canvasid", sm)
