"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var _ctx, _group, _circle, _text, _id, _name, _x, _y, _initial, _active, _rendered;
Object.defineProperty(exports, "__esModule", { value: true });
const fabric_1 = require("fabric");
const common_1 = require("./common");
// TODO: implement hover effect of state
// over
// target._objects[0].set('stroke', 'red');
//       target._objects[0].animate('radius', 32, {
//         duration: 300,
//         onChange: this.#canvas.renderAll.bind(this.#canvas),
//       });
// out
// const stroke = name === this.#initialState ? this.#initialStateStroke : this.#stateStroke;
// target._objects[0].set('stroke', stroke.toString());
// target._objects[0].animate('radius', 30, {
//   duration: 300,
//   onChange: this.#canvas.renderAll.bind(this.#canvas),
// });
class State {
    constructor(ctx, name, x, y, { initial = false, active = false } = {}) {
        _ctx.set(this, void 0);
        // canvas state
        _group.set(this, void 0);
        _circle.set(this, void 0);
        _text.set(this, void 0);
        // state properties
        _id.set(this, void 0);
        _name.set(this, void 0);
        _x.set(this, void 0);
        _y.set(this, void 0);
        _initial.set(this, void 0);
        _active.set(this, void 0);
        _rendered.set(this, false);
        __classPrivateFieldSet(this, _ctx, ctx);
        __classPrivateFieldSet(this, _id, common_1.genID());
        __classPrivateFieldSet(this, _name, name);
        __classPrivateFieldSet(this, _x, x);
        __classPrivateFieldSet(this, _y, y);
        __classPrivateFieldSet(this, _initial, initial);
        __classPrivateFieldSet(this, _active, active);
        const stroke = __classPrivateFieldGet(this, _initial) ? State.initialStateStroke : State.stateStroke;
        const fill = __classPrivateFieldGet(this, _active) ? State.activeStateFill : State.stateFill;
        __classPrivateFieldSet(this, _circle, new fabric_1.fabric.Circle({
            originX: 'center',
            originY: 'center',
            fill: fill.toString(),
            stroke: stroke.toString(),
            radius: State.stateRadius,
            shadow: new fabric_1.fabric.Shadow({
                color: 'rgba(0, 0, 0, 0.5)',
                offsetX: 5,
                offsetY: 5,
            }),
        }));
        __classPrivateFieldSet(this, _text, new fabric_1.fabric.Text(name, {
            originX: 'center',
            originY: 'center',
            fontSize: 10,
            stroke: State.textStroke.toString(),
        }));
        __classPrivateFieldSet(this, _group, new fabric_1.fabric.Group([__classPrivateFieldGet(this, _circle), __classPrivateFieldGet(this, _text)], {
            originX: 'center',
            originY: 'center',
            left: x,
            top: y,
            hasControls: false,
            // TODO: make this configurable
            selectable: true,
            padding: 2,
        }));
    }
    get name() {
        return __classPrivateFieldGet(this, _name);
    }
    get id() {
        return __classPrivateFieldGet(this, _id);
    }
    get position() {
        __classPrivateFieldSet(this, _x, __classPrivateFieldGet(this, _group).left || 0);
        __classPrivateFieldSet(this, _y, __classPrivateFieldGet(this, _group).top || 0);
        return { x: __classPrivateFieldGet(this, _x), y: __classPrivateFieldGet(this, _y) };
    }
    get initial() {
        return __classPrivateFieldGet(this, _initial);
    }
    set initial(value) {
        __classPrivateFieldSet(this, _initial, value);
        const stroke = __classPrivateFieldGet(this, _initial) ? State.initialStateStroke : State.stateStroke;
        __classPrivateFieldGet(this, _circle).set('stroke', stroke.toString());
        __classPrivateFieldGet(this, _ctx).renderAll();
    }
    get active() {
        return __classPrivateFieldGet(this, _active);
    }
    set active(value) {
        __classPrivateFieldSet(this, _initial, value);
        const fill = __classPrivateFieldGet(this, _active) ? State.activeStateFill : State.stateFill;
        __classPrivateFieldGet(this, _circle).set('fill', fill.toString());
        __classPrivateFieldGet(this, _ctx).renderAll();
    }
    on(event, handler) {
        __classPrivateFieldGet(this, _group).on(event, (opt) => handler(opt, this));
    }
    remove() {
        __classPrivateFieldGet(this, _ctx).remove(__classPrivateFieldGet(this, _group));
        __classPrivateFieldSet(this, _rendered, false);
    }
    render() {
        if (__classPrivateFieldGet(this, _rendered)) {
            this.remove();
        }
        __classPrivateFieldGet(this, _ctx).add(__classPrivateFieldGet(this, _group));
        __classPrivateFieldSet(this, _rendered, true);
    }
    setCoords() {
        __classPrivateFieldGet(this, _group).setCoords();
    }
}
exports.default = State;
_ctx = new WeakMap(), _group = new WeakMap(), _circle = new WeakMap(), _text = new WeakMap(), _id = new WeakMap(), _name = new WeakMap(), _x = new WeakMap(), _y = new WeakMap(), _initial = new WeakMap(), _active = new WeakMap(), _rendered = new WeakMap();
State.initialStateStroke = common_1.EDITOR_DEFAULTS.initialStateStroke;
State.activeStateFill = common_1.EDITOR_DEFAULTS.currentStateFill;
State.stateStroke = common_1.EDITOR_DEFAULTS.stateStroke;
State.stateFill = common_1.EDITOR_DEFAULTS.stateFill;
State.textStroke = common_1.EDITOR_DEFAULTS.textStroke;
State.stateRadius = common_1.EDITOR_DEFAULTS.stateRadius;
