"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _id, _name, _ctx, _link, _pointer, _src, _dest, _defaultX, _defaultY, _handlers;
Object.defineProperty(exports, "__esModule", { value: true });
const fabric_1 = require("fabric");
const state_1 = __importDefault(require("./state"));
const common_1 = require("./common");
class NewTransition {
    constructor(ctx, name, { src, dest, defaultX, defaultY } = {}) {
        _id.set(this, void 0);
        _name.set(this, void 0);
        _ctx.set(this, void 0);
        _link.set(this, void 0);
        _pointer.set(this, void 0);
        _src.set(this, void 0);
        _dest.set(this, void 0);
        _defaultX.set(this, void 0);
        _defaultY.set(this, void 0);
        _handlers.set(this, []);
        __classPrivateFieldSet(this, _id, common_1.genID());
        __classPrivateFieldSet(this, _name, name);
        __classPrivateFieldSet(this, _ctx, ctx);
        __classPrivateFieldSet(this, _src, src ?? null);
        __classPrivateFieldSet(this, _dest, dest ?? null);
        const { link, pointer } = this.render();
        __classPrivateFieldSet(this, _link, link);
        __classPrivateFieldSet(this, _pointer, pointer);
        __classPrivateFieldSet(this, _defaultX, defaultX || 0);
        __classPrivateFieldSet(this, _defaultY, defaultY || 0);
    }
    get src() {
        return __classPrivateFieldGet(this, _src);
    }
    set src(value) {
        __classPrivateFieldSet(this, _src, value);
        this.rerender();
    }
    get dest() {
        return __classPrivateFieldGet(this, _dest);
    }
    set dest(value) {
        __classPrivateFieldSet(this, _dest, value);
        this.rerender();
    }
    get name() {
        return __classPrivateFieldGet(this, _name);
    }
    get id() {
        return __classPrivateFieldGet(this, _id);
    }
    set defaultDest(value) {
        __classPrivateFieldSet(this, _defaultX, value.x);
        __classPrivateFieldSet(this, _defaultY, value.y);
        this.rerender();
    }
    get defaultDest() {
        return { x: __classPrivateFieldGet(this, _defaultX), y: __classPrivateFieldGet(this, _defaultY) };
    }
    render() {
        const defaultPos = { x: __classPrivateFieldGet(this, _defaultX), y: __classPrivateFieldGet(this, _defaultY) };
        const srcPos = __classPrivateFieldGet(this, _src)?.position ?? defaultPos;
        const destPos = __classPrivateFieldGet(this, _dest)?.position ?? defaultPos;
        const { link, pointer } = this.calcTransitionLink(srcPos.x, srcPos.y, destPos.x, destPos.y, 30);
        const stroke = common_1.mapText2Color(__classPrivateFieldGet(this, _name));
        __classPrivateFieldSet(this, _link, new fabric_1.fabric.Path(link, {
            fill: '',
            stroke,
            strokeWidth: 2,
            perPixelTargetFind: true,
            lockMovementX: true,
            lockMovementY: true,
            hasControls: false,
        }));
        __classPrivateFieldSet(this, _pointer, new fabric_1.fabric.Path(pointer, {
            stroke,
            fill: stroke,
            hasControls: false,
            selectable: false,
        }));
        __classPrivateFieldGet(this, _ctx).add(__classPrivateFieldGet(this, _link), __classPrivateFieldGet(this, _pointer));
        return { link: __classPrivateFieldGet(this, _link), pointer: __classPrivateFieldGet(this, _pointer) };
    }
    rerender() {
        this.remove();
        this.render();
        __classPrivateFieldGet(this, _handlers).forEach(({ event, handler }) => {
            __classPrivateFieldGet(this, _link).on(event, (opt) => handler(opt, this));
        });
    }
    remove() {
        __classPrivateFieldGet(this, _ctx).remove(__classPrivateFieldGet(this, _pointer));
        __classPrivateFieldGet(this, _ctx).remove(__classPrivateFieldGet(this, _link));
    }
    setCoords() {
        __classPrivateFieldGet(this, _pointer).setCoords();
        __classPrivateFieldGet(this, _link).setCoords();
    }
    on(event, handler) {
        __classPrivateFieldGet(this, _handlers).push({ event, handler });
        __classPrivateFieldGet(this, _link).on(event, (opt) => handler(opt, this));
    }
    // CALCULATIONS
    calcLinkPath(fromX, fromY, toX, toY, curve) {
        const midX = fromX + (toX - fromX) / 2;
        const midY = fromY + (toY - fromY) / 2;
        const pX = midX - curve;
        const pY = midY + curve;
        return `M ${fromX} ${fromY} Q ${pX}, ${pY}, ${toX}, ${toY}`;
    }
    calcPointerPath(x, y, rotation = 0) {
        const height = 10;
        const base = 10;
        const side = Math.sqrt((base / 2) ** 2 + height ** 2);
        const rad = Math.atan(base / 2 / height);
        const l1 = `L ${x + Math.cos(rad + rotation) * side} ${y + Math.sin(rad + rotation) * side}`;
        const l2 = `L ${x + Math.cos(-rad + rotation) * side} ${y + Math.sin(-rad + rotation) * side}`;
        return `M ${x} ${y} ${l1} ${l2} z`;
    }
    calcSelfLinkPath(x, y) {
        const bezierY = y - state_1.default.stateRadius;
        const y1 = bezierY - NewTransition.selfLinkHeight;
        const x2 = x - NewTransition.selfLinkHeight;
        const y2 = bezierY;
        return `M ${x} ${bezierY} C ${x} ${y1}, ${x2} ${y2}, ${x} ${bezierY}`;
    }
    calcTransitionLink(x1, y1, x2, y2, curve) {
        if (x1 === x2 && y1 === y2) {
            return {
                link: this.calcSelfLinkPath(x1, y1),
                pointer: this.calcPointerPath(x1, y1 - state_1.default.stateRadius, -Math.PI / 2),
            };
        }
        const theta = Math.atan((y2 - y1) / (x2 - x1));
        const sign = (theta > 0 ? -1 : 1) * (y2 > y1 ? -1 : 1);
        const fromX = x1 + sign * Math.cos(theta) * state_1.default.stateRadius;
        const fromY = y1 + sign * Math.sin(theta) * state_1.default.stateRadius;
        const toX = x2 + sign * -1 * Math.cos(theta) * state_1.default.stateRadius;
        const toY = y2 + sign * -1 * Math.sin(theta) * state_1.default.stateRadius;
        const midX = fromX + (toX - fromX) / 2;
        const midY = fromY + (toY - fromY) / 2;
        const pX = midX - curve * sign;
        const pY = midY + curve * sign;
        const pointerTheta = Math.atan((toY - pY) / (toX - pX));
        return {
            link: this.calcLinkPath(fromX, fromY, toX, toY, curve * sign),
            pointer: this.calcPointerPath(toX, toY, sign === -1 ? pointerTheta : Math.PI + pointerTheta),
        };
    }
}
exports.default = NewTransition;
_id = new WeakMap(), _name = new WeakMap(), _ctx = new WeakMap(), _link = new WeakMap(), _pointer = new WeakMap(), _src = new WeakMap(), _dest = new WeakMap(), _defaultX = new WeakMap(), _defaultY = new WeakMap(), _handlers = new WeakMap();
NewTransition.selfLinkHeight = common_1.EDITOR_DEFAULTS.selfLinkHeight;
